"""This module implements tools that are generally useful in dantro"""

import collections
import contextlib
import glob
import logging
import os
import sys
from datetime import timedelta as _timedelta
from shutil import get_terminal_size as _get_terminal_size
from typing import List, Mapping, Optional, Sequence, Set, Tuple, Union

import numpy as np

log = logging.getLogger(__name__)

# -- Terminal, TTY ------------------------------------------------------------

TERMINAL_INFO = dict(columns=79, lines=24, is_a_tty=False)
"""Holds information about the size and properties of the used terminal.

.. warning::

    Do not update this manually, call :py:func:`.update_terminal_info` instead.
"""


def update_terminal_info() -> dict:
    """Updates the ``TERMINAL_INFO`` constant with information about the
    number of columns, lines, and whether the terminal is a TTY terminal.

    If retrieving the properties via :py:func:`shutil.get_terminal_size` fails
    for whatever reason, will not apply any changes.
    """
    try:
        terminal_size = _get_terminal_size()
        cols, lines = terminal_size.columns, terminal_size.lines

    except Exception as exc:
        log.debug(
            "Failed to update terminal info!  %s: %s", type(exc).__name__, exc
        )

    else:
        TERMINAL_INFO["columns"] = cols
        TERMINAL_INFO["lines"] = lines
        TERMINAL_INFO["is_a_tty"] = sys.stdout.isatty()
        log.debug("Updated terminal info:  %s", TERMINAL_INFO)

    return TERMINAL_INFO


# Set the content and the (not updateable constants)
update_terminal_info()

IS_A_TTY = TERMINAL_INFO["is_a_tty"]
"""Whether the used terminal is a TTY terminal

.. deprecated:: v0.18

    Use the ``dantro.tools.TERMINAL_INFO["is_a_tty"]`` entry instead.
"""

TTY_COLS = TERMINAL_INFO["columns"]
"""Number of columns in a TTY terminal

.. deprecated:: v0.18

    Use the ``dantro.tools.TERMINAL_INFO["columns"]`` entry instead.
"""

# -- YAML ---------------------------------------------------------------------

from ._yaml import load_yml, write_yml, yaml

# -- Dictionary operations ----------------------------------------------------


def recursive_update(d: dict, u: dict) -> dict:
    """Recursively updates the Mapping-like object ``d`` with the Mapping-like
    object ``u`` and returns it. Note that this does *not* create a copy of
    ``d``, but changes it mutably!

    Based on: http://stackoverflow.com/a/32357112/1827608

    Args:
        d (dict): The mapping to update
        u (dict): The mapping whose values are used to update ``d``

    Returns:
        dict: The updated dict ``d``
    """
    for k, v in u.items():
        if isinstance(d, collections.abc.Mapping):
            # Already a Mapping
            if isinstance(v, collections.abc.Mapping):
                # Already a Mapping, continue recursion
                d[k] = recursive_update(d.get(k, {}), v)
                # This already creates a mapping if the key was not available
            else:
                # Not a mapping -> at leaf -> update value
                d[k] = v  # ... which is just u[k]

        else:
            # Not a mapping -> create one
            d = {k: u[k]}
    return d


def recursive_getitem(obj: Union[Mapping, Sequence], keys: Sequence):
    """Go along the sequence of ``keys`` through ``obj`` and return the target
    item.

    Args:
        obj (Union[Mapping, Sequence]): The object to get the item from
        keys (Sequence): The sequence of keys to follow

    Returns:
        The target item from ``obj``, specified by ``keys``

    Raises:
        ValueError: If any index or key in the key sequence was not available
    """

    def handle_error(exc: Exception, *, key, keys, obj):
        raise ValueError(
            f"Invalid {'key' if isinstance(exc, KeyError) else 'index'} "
            f"'{key}' during recursive getitem of key sequence "
            f"{' -> '.join([repr(k) for k in keys])}! "
            f"{exc.__class__.__name__}: {exc} raised on the following "
            f"object:\n{obj}"
        ) from exc

    if len(keys) > 1:
        # Continue recursion
        try:
            return recursive_getitem(obj[keys[0]], keys=keys[1:])
        except (KeyError, IndexError) as err:
            handle_error(err, key=keys[0], keys=keys, obj=obj)

    # else: reached the end of the recursion
    try:
        return obj[keys[0]]
    except (KeyError, IndexError) as err:
        handle_error(err, key=keys[0], keys=keys, obj=obj)


# -- Terminal messaging -------------------------------------------------------


def clear_line(only_in_tty=True, break_if_not_tty=True):
    """Clears the current terminal line and resets the cursor to the first
    position using a POSIX command.

    Based on: https://stackoverflow.com/a/25105111/1827608

    Args:
        only_in_tty (bool, optional): If True (default) will only clear the
            line if the script is executed in a TTY
        break_if_not_tty (bool, optional): If True (default), will insert a
            line break if the script is not executed in a TTY
    """
    # Differentiate cases
    if (only_in_tty and TERMINAL_INFO["is_a_tty"]) or not only_in_tty:
        # Print the POSIX character
        print("\x1b[2K\r", end="")

    if break_if_not_tty and not TERMINAL_INFO["is_a_tty"]:
        # print linebreak (no flush)
        print("\n", end="")

    # flush manually (there might not have been a linebreak)
    sys.stdout.flush()


def fill_line(
    s: str,
    *,
    num_cols: int = None,
    fill_char: str = " ",
    align: str = "left",
) -> str:
    """Extends the given string such that it fills a whole line of `num_cols`
    columns.

    Args:
        s (str): The string to extend to a whole line
        num_cols (int, optional): The number of colums of the line; defaults to
            the number of terminal columns.
        fill_char (str, optional): The fill character
        align (str, optional): The alignment. Can be: 'left', 'right', 'center'
            or the one-letter equivalents.

    Returns:
        str: The string of length `num_cols`

    Raises:
        ValueError: For invalid `align` or `fill_char` argument
    """
    if num_cols is None:
        num_cols = TERMINAL_INFO["columns"]

    if len(fill_char) != 1:
        raise ValueError(
            "Argument `fill_char` needs to be string of length 1 but was: "
            + str(fill_char)
        )

    fill_str = fill_char * (num_cols - len(s))

    if align in ["left", "l", None]:
        return s + fill_str

    elif align in ["right", "r"]:
        return fill_str + s

    elif align in ["center", "centre", "c"]:
        return (
            fill_str[: len(fill_str) // 2] + s + fill_str[len(fill_str) // 2 :]
        )

    raise ValueError(f"align argument '{align}' not supported")


def print_line(s: str, *, end="\r", **kwargs):
    """Wrapper around :py:func:`~dantro.tools.fill_line` that also prints
    a line with carriage return (without new line) as end character. This is
    useful for progress report lines that overwrite the previously printed
    content repetitively.
    """
    print(fill_line(s, **kwargs), end=end)


def center_in_line(
    s: str, *, num_cols: int = None, fill_char: str = "·", spacing: int = 1
) -> str:
    """Shortcut for a common fill_line use case.

    Args:
        s (str): The string to center in the line
        num_cols (int, optional): The number of columns in the line,
            automatically determined if not given
        fill_char (str, optional): The fill character
        spacing (int, optional): The spacing around the string `s`

    Returns:
        str: The string centered in the line
    """
    spacing = " " * spacing
    return fill_line(
        spacing + s + spacing,
        num_cols=num_cols,
        fill_char=fill_char,
        align="centre",
    )


def make_columns(
    items: List[str],
    *,
    wrap_width: int = None,
    fstr: str = "  {item:<{width:}s}  ",
) -> str:
    """Given a sequence of string items, returns a string with these items
    spread out over several columns. Iteration is first within the row and
    then into the next row.

    The number of columns is determined automatically from the wrap width, the
    length of the longest item in the items list, and the length of the
    evaluated format string.

    Args:
        items (List[str]): The string items to represent in columns.
        wrap_width (int, optional): The maximum width of each full row. If not
            given will determine it automatically
        fstr (str, optional): The format string to use. Needs to accept the
            keys ``item`` and ``width``, the latter of which will be used for
            padding. The format string should lead to strings of equal length,
            otherwise the column layout will be messed up!
    """
    if not items:
        return ""

    if not wrap_width:
        wrap_width = TERMINAL_INFO["columns"]

    max_item_width = max(len(item) for item in items)
    item_str_width = len(
        fstr.format(item=" " * max_item_width, width=max_item_width)
    )
    num_cols = wrap_width // item_str_width

    rows = []
    for i, item in enumerate(items):
        item_str = fstr.format(item=item, width=max_item_width)

        # New row or new column?
        if i % num_cols == 0:
            rows.append(item_str)
        else:
            rows[-1] += item_str

    return "\n".join(rows) + "\n"


# -- Fun with byte strings ----------------------------------------------------


def decode_bytestrings(obj) -> str:
    """Checks whether the given attribute value is or contains byte
    strings and if so, decodes it to a python string.

    Args:
        obj: The object to try to decode into holding python strings

    Returns:
        str: Either the unchanged object or the decoded one
    """
    # Check for data loaded as array of bytestring
    if isinstance(obj, np.ndarray):
        if obj.dtype.kind in ["S", "a"]:
            obj = obj.astype("U")

        # If it is of dtype object, decode all bytes objects
        if obj.dtype == np.dtype("object"):

            def decode_if_bytes(val):
                if isinstance(val, bytes):
                    return val.decode("utf8")
                return val

            # Apply element-wise
            obj = np.vectorize(decode_if_bytes)(obj)

    # ... or as bytes
    elif isinstance(obj, bytes):
        # Decode bytestring to unicode
        obj = obj.decode("utf8")

    return obj


# -- Misc ---------------------------------------------------------------------

DoNothingContext = contextlib.nullcontext
"""An alias for a context ... that does nothing"""


def ensure_dict(d: Optional[dict]) -> dict:
    """Makes sure that ``d`` is a dict and not None"""
    if d is None:
        return dict()
    return d


def is_iterable(obj) -> bool:
    """Tries whether the given object is iterable."""
    try:
        (e for e in obj)
    except:
        return False
    return True


def is_hashable(obj) -> bool:
    """Tries whether the given object is hashable."""
    try:
        hash(obj)
    except:
        return False
    return True


def try_conversion(c: str) -> Union[bool, int, float, complex, str, None]:
    """Given a string, attempts to convert it to a numerical value or a bool."""
    c = str(c)

    if c.lower() == "true":
        return True
    elif c.lower() == "false":
        return False
    elif c in ("~", "None", "none"):
        return None

    try:
        return int(c)
    except:
        pass

    try:
        return float(c)
    except:
        pass

    try:
        return complex(c)
    except:
        pass

    return c


def parse_str_to_args_and_kwargs(s: str, *, sep: str) -> Tuple[list, dict]:
    """Parses strings like ``65,0,sep=12`` into a positional arguments list
    and a keyword arguments dict.

    Behavior:

    * Positional arguments are all arguments that do *not* include ``=``.
      Keyword arguments are those that *do* include ``=``.
    * Will use :py:func:`.try_conversion` to convert argument values.
    * Trailing and leading white space on argument names and values is stripped
      away using :py:meth:`~str.strip`.

    .. warning::

        * Cannot handle string arguments that include ``sep`` or ``=``!
        * Cannot handle arguments that define lists, tuples or other more
          complex objects.

    .. hint::

        For more complex argument parsing, consider using a YAML parser
        instead of this (rather simple) function!
    """
    all_args = s.split(sep)
    args = [
        try_conversion(a.strip()) for a in all_args if (a and "=" not in a)
    ]
    kwargs = {
        k.strip(): try_conversion(v.strip())
        for k, v in [kw.split("=") for kw in all_args if "=" in kw]
    }
    return args, kwargs


class adjusted_log_levels:
    """A context manager that temporarily adjusts log levels"""

    def __init__(self, *new_levels: Sequence[Tuple[str, int]]):
        self.new_levels = {n: l for n, l in new_levels}
        self.old_levels = dict()

    def __enter__(self):
        """When entering the context, sets these levels"""
        for name, new_level in self.new_levels.items():
            logger = logging.getLogger(name)
            self.old_levels[name] = logger.level
            logger.setLevel(new_level)

    def __exit__(self, *_):
        """When leaving the context, resets the levels to their old state"""
        for name, old_level in self.old_levels.items():
            logging.getLogger(name).setLevel(old_level)


def total_bytesize(files: List[str]) -> int:
    """Returns the total size of a list of files"""
    return sum(os.path.getsize(fpath) for fpath in files)


def format_bytesize(num: int, *, precision: int = 1) -> str:
    """Formats a size in bytes to a human readable (binary) format.

    Stripped down from https://stackoverflow.com/a/63839503/1827608 .

    Args:
        num (int): Number of bytes
        precision (int, optional): The decimal precision to use, can be 0..3

    Returns:
        str: The formatted, human-readable byte size
    """
    UNIT_LABELS = ("B", "kiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    PRECISION_OFFSETS = (0.5, 0.05, 0.005, 0.0005)
    FSTRS = (
        "{}{:.0f} {}",
        "{}{:.1f} {}",
        "{}{:.2f} {}",
        "{}{:.3f} {}",
    )

    last_label = UNIT_LABELS[-1]
    unit_step = 1024
    unit_step_thresh = unit_step - PRECISION_OFFSETS[precision]
    is_negative = num < 0

    if is_negative:
        num = abs(num)

    # Special case for Bytes, where there should be no decimal point
    if num < 1024:
        return FSTRS[0].format("-" if is_negative else "", num, "B")

    for unit in UNIT_LABELS:
        if num < unit_step_thresh:
            # Below threshold now, can go to formatting
            break

        # Unless we reached the highest prefix, shrink the number such that
        # it represents the value in the next higher unit
        if unit != last_label:
            num /= unit_step

    return FSTRS[precision].format("-" if is_negative else "", num, unit)


def format_time(
    duration: Union[float, _timedelta],
    *,
    ms_precision: int = 0,
    max_num_parts: int = None,
) -> str:
    """Given a duration (in seconds), formats it into a string.

    The formatting divisors are: days, hours, minutes, seconds

    If ``ms_precision`` > 0 and ``duration`` < 60, decimal places will be shown
    for the seconds.

    Args:
        duration (Union[float, datetime.timedelta]): The duration in seconds
            to format into a duration string; it can also be a timedelta
            object.
        ms_precision (int, optional): The precision of the seconds slot
        max_num_parts (int, optional): How many parts to include when creating
            the formatted time string. For example, if the time consists of
            the parts seconds, minutes, and hours, and the argument is ``2``,
            only the hours and minutes parts will be shown, thus reducing the
            precision of the overall representation of ``duration``.
            If None, all parts are included.

    Returns:
        str: The formatted duration string
    """

    if isinstance(duration, _timedelta):
        duration = duration.total_seconds()

    divisors = (24 * 60 * 60, 60 * 60, 60, 1)
    letters = ("d", "h", "m", "s")
    remaining = float(duration)
    parts = []

    # Also handle negative numbers
    is_negative = bool(duration < 0)
    if is_negative:
        # Calculate with the positive value, much easier
        remaining *= -1

    # Go over divisors and letters and see if there is something to represent
    for divisor, letter in zip(divisors, letters):
        time_to_represent = int(remaining / divisor)
        remaining -= time_to_represent * divisor

        if time_to_represent > 0:
            # Distinguish between seconds and other divisors for short times
            if ms_precision <= 0 or abs(duration) >= 60:
                # Regular behaviour: Seconds do not have decimals or duration
                # is so long that they need not be represented.
                s = f"{time_to_represent:d}{letter}"

            else:
                # There are decimals to be represented.
                s = "{val:.{prec:d}f}s".format(
                    val=(time_to_represent + remaining),
                    prec=int(ms_precision),
                )

            parts.append(s)

    # If nothing was added so far, the time was below one second
    if not parts:
        if duration == 0:
            return "0s"

        elif ms_precision == 0:
            # Just show an approximation
            if not is_negative:
                return "< 1s"
            return "> -1s"

        # Show with ms_precision decimal places
        s = "{val:{tot}.{prec}f}s".format(
            val=remaining,
            tot=int(ms_precision) + 2,
            prec=int(ms_precision),
        )
        parts.append(s)

    # May need to prepend a minus for negative values
    if is_negative:
        parts = ["-"] + parts

    # Join the parts together, but only the maximum number of parts
    if not max_num_parts:
        return " ".join(parts)
    return " ".join(parts[: max_num_parts + int(is_negative)])


def glob_paths(
    glob_str: Union[str, List[str]],
    *,
    ignore: List[str] = None,
    base_path: str = None,
    sort: bool = False,
    recursive: bool = True,
    include_files: bool = True,
    include_directories: bool = True,
) -> List[str]:
    """Generates a list of paths from a glob string and a number of additional
    options.

    Paths may refer to file *and* directory paths.
    Uses :py:func:`glob.glob` for matching glob strings.

    .. note::

        Internally, this uses a set, thus ensuring that there are no duplicate
        paths in the returned list.

    Args:
        glob_str (Union[str, List[str]]): The glob pattern or a list of
            glob patterns to use for searching for files. Relative paths will
            be seen as relative to ``base_path``.
        ignore (List[str]): A list of paths to ignore. Relative paths will be
            seen as relative to ``base_path``. Supports glob patterns.
        base_path (str, optional): The base path for the glob pattern. If not
            given, will use the current working directory.
        sort (bool, optional): If true, sorts the list before returning.
        recursive (bool, optional): If true, will activate recursive glob
            patterns (see :py:func:`glob.glob`).
        include_files (bool, optional): If false, will remove file paths from
            the set of paths.
        include_directories (bool, optional): If false, will remove directory
            paths from the set of paths.

    Returns:
        List[str]:
            The file or directory paths that matched ``glob_str`` and were not
            filtered out by the other options.

    Raises:
        ValueError:
            If the given ``base_path`` was not absolute.
    """

    def prepare_path(path: str, *, base: str) -> str:
        return os.path.abspath(os.path.join(base, os.path.expanduser(path)))

    def remove_from_set(s: set, to_remove: str):
        try:
            s.remove(to_remove)
        except KeyError:
            log.debug("%s was not found in set of paths.", to_remove)
        else:
            log.debug("%s removed from set of paths.", to_remove)

    # Create a set to assure that there are no duplicate entries
    paths = set()

    # Assure it is a list of strings
    if isinstance(glob_str, str):
        glob_str = [glob_str]

    log.debug(
        "Got %d glob string(s) to create set of matching file paths from.",
        len(glob_str),
    )

    # Handle base path, defaulting to the data directory
    if base_path is None:
        base_path = os.getcwd()
        log.debug("Using current working directory as base path.")

    else:
        if not os.path.isabs(base_path):
            raise ValueError(
                "Given base_path argument needs be an "
                f"absolute path, was not: {base_path}"
            )

    # Go over the given glob strings and add to the paths set
    for gs in glob_str:
        # Make the glob string absolute and prepend the base path
        gs = prepare_path(gs, base=base_path)
        log.debug("Adding paths that match glob string:\n  %s", gs)

        # Add to the set of paths; this assures uniqueness of the paths
        paths.update(list(glob.glob(gs, recursive=recursive)))

    # See if some paths should be ignored
    ignore = ignore if ignore else []
    for igs in ignore:
        igs = prepare_path(igs, base=base_path)
        log.debug("Removing paths that match ignore glob string:\n  %s", igs)

        for ignore_path in glob.glob(igs, recursive=recursive):
            remove_from_set(paths, to_remove=ignore_path)

    # Can convert to list now, easier to continue
    paths = list(paths)

    # Finish up: May want to filter files / directories, and sort the list
    if not include_files:
        paths = [path for path in paths if not os.path.isfile(path)]

    if not include_directories:
        paths = [path for path in paths if not os.path.isdir(path)]

    if sort:
        paths.sort()

    return paths


# -- Multi-processing ---------------------------------------------------------


class PoolCallbackHandler:
    """A simple callback handler for multiprocessing pools"""

    def __init__(
        self,
        n_max: int,
        *,
        silent: bool = False,
        fstr: str = "  Loaded  {n}/{n_max} .",
    ):
        """
        Args:
            n_max (int): Number of tasks
            silent (bool, optional): If true, will *not* print a message
            fstr (str, optional): The format string for the status message.
                May contain keys ``n`` and ``n_max``.
        """
        self._n = 0
        self._n_max = n_max
        self.silent = silent
        self._fstr = fstr

    def __call__(self, _):
        self._n += 1
        if not self.silent:
            print_line(self._fstr.format(n=self._n, n_max=self._n_max))


class PoolErrorCallbackHandler:
    """A simple callback handler for errors in multiprocessing pools"""

    def __init__(self):
        self._errors = set()

    def __call__(self, error: Exception):
        self.track_error(error)

    def track_error(self, error: Exception):
        self._errors.update({error})

    @property
    def errors(self) -> Set[Exception]:
        return self._errors



"""Implements operations that need to be importable from other modules"""

import operator

# -----------------------------------------------------------------------------

# fmt: off
BOOLEAN_OPERATORS = {
    "==": operator.eq,  "eq": operator.eq,
    "<":  operator.lt,  "lt": operator.lt,
    "<=": operator.le,  "le": operator.le,
    ">":  operator.gt,  "gt": operator.gt,
    ">=": operator.ge,  "ge": operator.ge,
    "!=": operator.ne,  "ne": operator.ne,
    "^":  operator.xor, "xor": operator.xor,
    #
    # Expecting an iterable as second argument
    "contains":         operator.contains,
    "in":               (lambda x, y: x in y),
    "not in":           (lambda x, y: x not in y),
    #
    # Performing bitwise boolean operations to support numpy logic
    "in interval":      (lambda x, y: x >= y[0] & x <= y[1]),
    "not in interval":  (lambda x, y: x < y[0] | x > y[1]),
} # End of boolean operator definitions
"""Boolean binary operators"""
# fmt: on



"""Custom, optimized copying functions used thoughout dantro"""

import copy
import pickle as _pickle
from typing import Any

# -----------------------------------------------------------------------------

_shallowcopy = copy.copy
"""An alias for a shallow copy function used throughout dantro, currently
pointing to :py:func:`copy.copy`.
"""


def _deepcopy(obj: Any) -> Any:
    """A pickle-based deep-copy overload, that uses :py:func:`copy.deepcopy`
    only as a fallback option if serialization was not possible.

    Calls :py:func:`pickle.loads` on the output of :py:func:`pickle.dumps` of
    the given object.

    The pickling approach being based on a C implementation, this can easily
    be many times faster than the pure-Python-based :py:func:`copy.deepcopy`.
    """
    try:
        return _pickle.loads(_pickle.dumps(obj))
    except:
        return copy.deepcopy(obj)



"""Implements :py:class:`~dantro.base.BaseDataContainer` specializations.

isort:skip_file
"""

# To avoid circular imports, need to import this first
from ..base import BaseDataContainer

from ._registry import register_container, is_container, CONTAINERS
from .general import (
    MutableMappingContainer,
    MutableSequenceContainer,
    ObjectContainer,
    PassthroughContainer,
    StringContainer,
)
from .link import LinkContainer
from .numeric import NumpyDataContainer
from .path import PathContainer
from .xr import XrDataContainer



"""This sub-package provides mixin classes for easily adding functionality
to a derived contaier or group class
"""

from .base import (
    AttrsMixin,
    BasicComparisonMixin,
    CheckDataMixin,
    CollectionMixin,
    DirectInsertionModeMixin,
    ItemAccessMixin,
    LockDataMixin,
    MappingAccessMixin,
    SizeOfMixin,
)
from .general import ForwardAttrsMixin, ForwardAttrsToDataMixin
from .indexing import IntegerItemAccessMixin, PaddedIntegerItemAccessMixin
from .numeric import ComparisonMixin, NumbersMixin, UnaryOperationsMixin
from .proxy_support import Hdf5ProxySupportMixin, ProxySupportMixin



"""This module implements a deterministic hash function to use within dantro.

It is mainly used for all things related to the TransformationDAG.
"""

import hashlib

# The full hash length of an MD5 hexdigest
FULL_HASH_LENGTH = 32

# The length of shortened hashes
SHORT_HASH_LENGTH = 8  # For hexdigest, 16^8 = 2^32 > 4*10^9 -> plenty.


def _hash(s: str) -> str:
    """Returns a deterministic hash of the given string.

    This uses the hashlib.md5 algorithm which returns a hexadecimal digest of
    length 32.

    .. note::

        This hash is meant to be used as a checksum, not for security.

    Args:
        s (str): The string to create the hash of

    Returns:
        str: The 32 character hexadecimal md5 hash digest
    """
    try:
        return hashlib.md5(s.encode("utf-8")).hexdigest()

    except Exception:
        # Probably was already a byte string
        return hashlib.md5(s).hexdigest()



"""Defines a data loader for Python pickles."""

from .._import_tools import LazyLoader
from ..containers import ObjectContainer
from ._registry import add_loader

pkl = LazyLoader("dill")

# -----------------------------------------------------------------------------


class PickleLoaderMixin:
    """Supplies a load function for pickled python objects.

    For unpickling, the ``dill`` package is used.
    """

    @add_loader(TargetCls=ObjectContainer, register_aliases=["pkl"])
    def _load_pickle(
        filepath: str, *, TargetCls: type, **pkl_kwargs
    ) -> ObjectContainer:
        """Load a pickled object using :py:func:`dill._dill.load`.

        Args:
            filepath (str): Where the pickle-dumped file is located
            TargetCls (type): The class constructor
            **pkl_kwargs: Passed on to :py:func:`dill._dill.load`

        Returns:
            ObjectContainer: The unpickled object, stored in a dantro container
        """
        with open(filepath, mode="rb") as f:
            obj = pkl.load(f, **pkl_kwargs)

        # Populate the target container with the object
        return TargetCls(data=obj, attrs=dict(filepath=filepath))



"""Plot functions to draw :py:class:`networkx.Graph` objects.

.. todo::

    Should really integrate utopya ``GraphPlot`` here!
"""

import copy
import logging
import os
from typing import Callable, Union

from ...tools import recursive_update as _recursive_update

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def _wiggle_pos(
    pos: dict, *, x: float = None, y: float = None, seed: int = None
) -> dict:
    """Wiggles positions by absolute random amplitudes in x and y direction

    Args:
        pos (dict): Positions dict with values being x and y positions
        x (float, optional): Absolute wiggle amplitude
        y (float, optional): Absolute wiggle amplitude
        seed (int, optional): Seed for the :py:class:`numpy.random.RandomState`
            that is used for drawing random numbers. Set to a fixed value to
            always get the same positions.
    """
    import numpy as np
    import numpy.random

    rng = np.random.RandomState(seed)
    wiggle = lambda old_pos, amp: old_pos + amp * rng.uniform(-1, +1)

    if x is not None:
        pos = {n: [wiggle(p[0], x), p[1]] for n, p in pos.items()}

    if y is not None:
        pos = {n: [p[0], wiggle(p[1], y)] for n, p in pos.items()}

    return pos


def _get_positions(
    g: "networkx.Graph",
    *,
    model: Union[str, Callable],
    wiggle: dict = None,
    **kwargs,
) -> dict:
    """Returns the positions dict for the given graph, created from a networkx
    layouting algorithm of a certain name or an arbitrary callable.

    Args:
        g (networkx.Graph): The graph object for which to create the layout
        model (Union[str, Callable]): Name of the layouting model or the
            layouting function itself.
            If starting with ``graphviz_<prog>``, will invoke
            :py:func:`networkx.drawing.nx_agraph.graphviz_layout` with the
            given value for ``prog``. Note that these only take a single
            keyword argument, ``args``.
            If it is a string, it's looked up from the networkx namespace.
            If it is a callable, it is invoked with ``g`` as only positional
            argument and ``**kwargs`` as keyword arguments.
        wiggle (dict, optional): If given, will postprocess the positions dict
            by randomly wiggling x and y coordinates according to the absolute
            amplitudes given as values.
        **kwargs: Passed on to the layouting algorithm.
    """

    import networkx as nx

    if callable(model):
        log.debug("Invoking callable for node layouting ...")
        pos = model(g, **kwargs)

    elif model.startswith("graphviz_"):
        log.debug("Invoking %s model for node layouting ...", model)
        try:
            model = model[len("graphviz_") :]
            pos = nx.drawing.nx_agraph.graphviz_layout(g, prog=model, **kwargs)

        except ImportError as err:
            raise ImportError(
                "Could not apply graphviz layout, probably because "
                "pygraphviz is not installed!"
            ) from err

    else:
        _suffix = "_layout"
        POSITIONING_MODELS_NETWORKX = {
            l[: -len(_suffix)]: getattr(nx, l)
            for l in dir(nx)
            if l.endswith(_suffix)
        }

        try:
            log.debug("Invoking %s model for node layouting ...", model)
            layout_func = POSITIONING_MODELS_NETWORKX[model]
            pos = layout_func(g, **kwargs)

        except KeyError as err:
            _avail = ", ".join(POSITIONING_MODELS_NETWORKX)
            raise ValueError(
                f"No layouting model '{model}' available in networkx! "
                f"Available models: {_avail}"
            ) from err

    if wiggle:
        pos = _wiggle_pos(pos, **wiggle)
    return pos


def get_positions(
    g: "networkx.Graph",
    *,
    model: Union[str, Callable] = "spring",
    model_kwargs: dict = {},
    fallback: Union[str, dict] = None,
    silent_fallback: bool = False,
    **kwargs,
) -> dict:
    """Returns the positions dict for the given graph, created from a networkx
    layouting algorithm of a certain name or an arbitrary callable.

    This is a wrapper around :py:func:`._get_positions` which allows to specify
    a fallback layouting model to use in case the first one fails for whatever
    reason.

    Args:
        g (networkx.Graph): The graph object for which to create the layout
        model (Union[str, Callable], optional): Name of the layouting model or
            the layouting function itself.
            If starting with ``graphviz_<prog>``, will invoke
            :py:func:`networkx.drawing.nx_agraph.graphviz_layout` with the
            given value for ``prog``. Note that these only take a single
            keyword argument, ``args``.
            If it is a string, it's looked up from the networkx namespace.
            If it is a callable, it is invoked with ``g`` as only positional
            argument and ``**kwargs`` as keyword arguments.
        model_kwargs (dict, optional): A dict where keys correspond to names
            of layouting models and values are parameters that are to be passed
            to the layouting function. This dict may contain more arguments
            than required, only the ``model`` key is looked up here. This can
            be useful for providing a wider set of defaults. These defaults
            are not considered when ``model`` is a callable.
        fallback (Union[str, dict], optional): The fallback model name (if
            a string) or a dict containing the key ``model`` and further
            kwargs.
        silent_fallback (bool, optional): Whether to log a visible message
            about the fallback or a more discrete one.
        **kwargs: Passed on to the layouting algorithm in addition to the
            selected entry from ``model_kwargs``. Keys given here update those
            from ``model_kwargs``.
            Also, these are *not* passed on to the fallback invocation.
    """

    def parse_kwargs(model: Union[str, Callable], **kwargs) -> dict:
        """Performs lookup and update of the layouting model arguments"""
        if callable(model):
            return kwargs

        return _recursive_update(
            copy.deepcopy(model_kwargs.get(model, {})), kwargs
        )

    # Prepare arguments
    if isinstance(fallback, str):
        fallback = dict(model=fallback)

    # Get the positions, potentially using a fallback
    try:
        return _get_positions(g, model=model, **parse_kwargs(model, **kwargs))

    except Exception as exc:
        if not fallback:
            raise

        if not silent_fallback:
            log.caution(
                "Node layouting with '%s' model failed with a %s: %s",
                model,
                type(exc).__name__,
                exc,
            )
            log.remark("Invoking fallback layouting:  %s", fallback)
        else:
            log.remark(
                "Node layouting with '%s' model failed; using fallback:  %s",
                model,
                fallback,
            )

    return _get_positions(
        g, model=fallback["model"], **parse_kwargs(**fallback)
    )


# .............................................................................


def _draw_graph(
    g: "networkx.Graph",
    *,
    ax: "matplotlib.axes.Axes" = None,
    drawing: dict = {},
    layout: dict = {},
) -> list:
    """Draws a graph using
    :py:func:`networkx.drawing.nx_pylab.draw_networkx_nodes`,
    :py:func:`networkx.drawing.nx_pylab.draw_networkx_edges`, and
    :py:func:`networkx.drawing.nx_pylab.draw_networkx_labels`.

    .. warning::

        This function is not yet completed and may change anytime.

    Args:
        g (networkx.Graph): The graph to draw
        out_path (str): Where to store it to
        drawing (dict, optional): Drawing arguments, containing the
            ``nodes``, ``edges`` and ``labels`` keys. The ``labels`` key
            can contain the ``from_attr`` key which will read the attribute
            specified there and use it for the label.
        layout (dict, optional): Used to generate node positions via the
            :py:func:`~dantro.plot.funcs.graph.get_positions` function.
    """
    import matplotlib.pyplot as plt
    import networkx as nx

    def draw(
        g, *, ax, pos, nodes: dict = {}, edges: dict = {}, labels: dict = {}
    ):
        # Parse some attributes
        if "from_attr" in labels:
            labels["labels"] = nx.get_node_attributes(
                g, labels.pop("from_attr")
            )

        # Now draw
        _nodes = nx.draw_networkx_nodes(g, pos=pos, ax=ax, **nodes)
        _edges = nx.draw_networkx_edges(g, pos=pos, ax=ax, **edges)
        _labels = nx.draw_networkx_labels(g, pos=pos, ax=ax, **labels)

        # Gather artists, adapting to the specific way they are passed back
        artists = []
        artists.append(_nodes)
        artists += _edges
        artists += list(_labels.values())

        return artists

    ax = ax if ax is not None else plt.gca()
    pos = get_positions(g, **layout)

    # Draw
    artists = draw(g, ax=ax, pos=pos, **drawing)

    # Post-process
    ax.axis("off")

    return artists



#!/usr/bin/env python3
# fmt: off

from setuptools import find_packages, setup

# Dependencies for dantro itself
install_deps = [
    "numpy",
    "xarray >= 2022.11",
    "dask",
    "toolz",             # For dask.delayed
    "distributed",       # For dask's distributed scheduler
    "scipy",             # Used as a netcdf4 storage engine for xarray
    "sympy",
    "h5py",
    "h5netcdf",          # A more capable, h5py-based netcdf backend
    "matplotlib",
    "seaborn",
    "networkx",
    "ruamel.yaml",
    "dill",              # For faster and more powerful pickling
    "paramspace",
]
# NOTE When changing any of the dependencies, make sure to update the table of
#      dependencies in README.md.
#      When adding a NEW dependency, make sure to denote it in the isort
#      configuration, see pyproject.toml.


# Dependencies for running tests and general development of dantro
test_deps = [
    "pytest",
    "pytest-cov",
    "tox",
    "pre-commit",
    #
    # dependencies that expand xarray functionality
    "pooch",
    "cftime",
]

# Dependencies for building the dantro documentation
doc_deps = [
    "sphinx==4.*",
    "sphinx-book-theme>=0.3.*",
    "sphinx-togglebutton",
    "ipython",
]

# .............................................................................

DESCRIPTION = "Handle, transform, and visualize hierarchically structured data"
LONG_DESCRIPTION = """
``dantro``: handle, transform, and visualize hierarchically structured data
===========================================================================

``dantro`` – from *data* and *dentro* (Greek for *tree*) – is a Python
package that provides a uniform interface for hierarchically structured
and semantically heterogeneous data. It is built around three main
features:

-  **data handling:** loading heterogeneous data into a tree-like data
   structure, providing a uniform interface to it
-  **data transformation:** performing arbitrary operations on the data,
   if necessary using lazy evaluation
-  **data visualization:** creating a visual representation of the
   processed data

Together, these stages constitute a **data processing pipeline**: an
automated sequence of predefined, configurable operations. Akin to a
Continuous Integration pipeline, a data processing pipeline provides a
uniform, consistent, and easily extensible infrastructure that
contributes to more efficient and reproducible workflows. This can be
beneficial especially in a scientific context, for instance when
handling data that was generated by computer simulations.

``dantro`` is meant to be *integrated* into projects and be used to set up
such a data processing pipeline, customized to the needs of the project.
It is designed to be **easily customizable** to the requirements of the project
it is integrated in, even if the involved data is hierachically structured or
semantically heterogeneous.
Furthermore, it allows a **configuration-based specification** of all
operations via `YAML <https://en.wikipedia.org/wiki/YAML>`_ configuration
files; the resulting pipeline can then be controlled entirely via these
configuration files and without requiring code changes.

The ``dantro`` package is **open source software** released under the
`LGPLv3+ <https://www.gnu.org/licenses/lgpl-3.0.html>`_ license.
It was developed alongside the `Utopia project <https://gitlab.com/utopia-project/utopia>`_
(a modelling framework for complex and adaptive systems), but is an
independent package.

Learn more
----------

* `Documentation <https://dantro.readthedocs.io/>`_
* `Project page <https://gitlab.com/utopia-project/dantro>`_
* `README and installation instructions <https://gitlab.com/utopia-project/dantro#installing-dantro>`_
* `Publication in the Journal of Open Source Software <https://doi.org/10.21105/joss.02316>`_
* `Utopia Project Website <https://utopia-project.org/>`_

"""


# .............................................................................
# fmt: on

# A function to extract version number from __init__.py
def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import codecs
    import os
    import re

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................

setup(
    name="dantro",
    #
    # Set the version from dantro.__version__
    version=find_version("dantro", "__init__.py"),
    #
    # Project info
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="dantro developers",
    author_email="dantro-dev@iup.uni.heidelberg.de",
    license="LGPL-3.0-or-later",
    url="https://gitlab.com/utopia-project/dantro",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or"
        " later (LGPLv3+)",
    ],
    #
    # Distribution details, dependencies, ...
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_data=dict(dantro=["cfg/*.yml"]),
    data_files=[("", ["COPYING", "COPYING.LESSER", "README.md"])],
    python_requires=">=3.8",
    install_requires=install_deps,
    tests_require=test_deps,
    test_suite="py.test",
    extras_require=dict(
        test=test_deps,
        doc=doc_deps,
        dev=test_deps + doc_deps,
    ),
)



"""This module holds the data operations database"""

import logging
import math
import operator
from typing import Any, Callable, Dict, Iterable, List, Sequence, Tuple, Union

import numpy as np

from .._import_tools import (
    LazyLoader,
    get_from_module,
    import_module_or_object,
)
from ..exceptions import *
from ..tools import recursive_getitem
from ..utils.coords import extract_coords_from_attrs, extract_dim_names
from ..utils.ordereddict import KeyOrderedDict
from .arr_ops import *
from .ctrl_ops import *
from .expr_ops import *

log = logging.getLogger(__name__)

xr = LazyLoader("xarray")
nx = LazyLoader("networkx")
pd = LazyLoader("pandas")
scipy = LazyLoader("scipy")


# -----------------------------------------------------------------------------
# -- The Operations Database --------------------------------------------------
# -----------------------------------------------------------------------------

# fmt: off
# NOTE If a single "object to act upon" can be reasonably defined for an
#      operation, it should be accepted as the first positional argument.
_OPERATIONS = KeyOrderedDict({
    # General operations - - - - - - - - - - - - - - - - - - - - - - - - - - -
    "define":               lambda d: d,
    "pass":                 lambda d: d,

    # Functions useful for debugging
    "print":                print_data,

    # Control flow functions, e.g. conditionally skipping a plot
    "raise_SkipPlot":       raise_SkipPlot,

    # Working on imported modules (useful if other operations don't match)
    "from_module":          get_from_module,
    "import":               import_module_or_object,
    "call":                 lambda c, *a, **k: c(*a, **k),
    "import_and_call":
        lambda m, n, *a, **k: import_module_or_object(m, n)(*a, **k),

    # Defining and calling lambdas
    "lambda":               generate_lambda,
    "call_lambda":          lambda e, *a, **k: generate_lambda(e)(*a, **k),

    # Import from packages and call directly
    "math.":
        lambda ms, *a, **k: get_from_module(math, name=ms)(*a, **k),
    "np.":
        lambda ms, *a, **k: get_from_module(np, name=ms)(*a, **k),
    "xr.":
        lambda ms, *a, **k: get_from_module(xr, name=ms)(*a, **k),
    "scipy.":
        lambda ms, *a, **k: get_from_module(scipy, name=ms)(*a, **k),
    "nx.":
        lambda ms, *a, **k: get_from_module(nx, name=ms)(*a, **k),
    "pd.":
        lambda ms, *a, **k: get_from_module(pd, name=ms)(*a, **k),

    # Some commonly used types
    "list":                 list,
    "dict":                 dict,
    "tuple":                tuple,
    "set":                  set,

    "int":                  int,
    "float":                float,
    "complex":              complex,
    "bool":                 bool,
    "str":                  str,

    # Item access and manipulation
    "[]":                   lambda d, k:    d[k],
    "getitem":              lambda d, k:    d[k],
    "setitem":              lambda d, k, v: d.__setitem__(k, v),
    "recursive_getitem":    recursive_getitem,

    # Attribute-related
    ".":                    getattr,
    "getattr":              getattr,
    "setattr":              setattr,
    ".()":                  lambda d, attr, *a, **k: getattr(d, attr)(*a, **k),
    "callattr":             lambda d, attr, *a, **k: getattr(d, attr)(*a, **k),

    # Other common Python builtins
    "all":                  all,
    "any":                  any,
    "len":                  len,
    "min":                  min,
    "max":                  max,
    "sum":                  sum,
    "map":                  map,
    "repr":                 repr,
    "sorted":               sorted,

    # Common operations on strings
    ".format":              lambda s, *a, **k: s.format(*a, **k),
    ".join":                lambda s, *a, **k: s.join(*a, **k),
    ".split":               lambda s, *a, **k: s.split(*a, **k),


    # Numerical operations - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Broadly categorized in how many arguments they accept ...
    # Unary ...................................................................
    "neg":                  operator.neg,
    "pos":                  operator.pos,
    "truth":                operator.truth,
    "invert":               operator.invert,

    "increment":            lambda d: d + 1,
    "decrement":            lambda d: d - 1,
    "count_unique":         count_unique,

    # numpy
    ".T":                   lambda d: d.T,
    ".any":                 lambda d: d.any(),
    ".all":                 lambda d: d.all(),
    ".dtype":               lambda d: d.dtype,
    ".shape":               lambda d: d.shape,
    ".ndim":                lambda d: d.ndim,
    ".size":                lambda d: d.size,
    ".itemsize":            lambda d: d.itemsize,
    ".nbytes":              lambda d: d.nbytes,
    ".base":                lambda d: d.base,
    ".imag":                lambda d: d.imag,
    ".real":                lambda d: d.real,
    ".nonzero":             lambda d: d.nonzero,
    ".flat":                lambda d: d.flat,
    ".item":                lambda d: d.item(),

    # xarray
    ".data":                lambda d: d.data,
    ".values":              lambda d: d.values,
    ".name":                lambda d: d.name,
    ".head":                lambda d: d.head(),
    ".tail":                lambda d: d.tail(),
    ".isnull":              lambda d: d.isnull(),

    # logarithms and powers, all numpy-based
    "log":                  np.log,
    "log10":                np.log10,
    "log2":                 np.log2,
    "log1p":                np.log1p,
    "squared":              np.square,
    "sqrt":                 np.sqrt,
    "cubed":                lambda d: np.power(d, 3),
    "sqrt3":                lambda d: np.power(d, 1./3.),

    # normalization and cumulation
    "normalize_to_sum":     lambda d: d / np.sum(d),
    "normalize_to_max":     lambda d: d / np.max(d),
    "cumulate":             lambda d, **k: np.cumsum(d, **k),
    "cumulate_complementary": lambda d, **k: np.cumsum(d[::-1], **k)[::-1],


    # Binary ..................................................................
    # Elementwise operations
    "add":                  operator.add,
    "concat":               operator.concat,
    "div":                  operator.truediv,
    "truediv":              operator.truediv,
    "floordiv":             operator.floordiv,
    "lshift":               operator.lshift,
    "mod":                  operator.mod,
    "mul":                  operator.mul,
    "matmul":               operator.matmul,
    "pow":                  operator.pow,
    "rshift":               operator.rshift,
    "sub":                  operator.sub,

    # numpy
    "np.power":             np.power,
    "np.dot":               np.dot,

    # xarray
    ".coords":
        lambda d, k=None: d.coords if k is None else d.coords[k],
    ".attrs":
        lambda d, k=None: d.attrs if k is None else d.attrs[k],
    ".variables":
        lambda d, k=None: d.variables if k is None else d.variables[k],
    ".data_vars":
        lambda d, k=None: d.data_vars if k is None else d.data_vars[k],


    # N-ary ...................................................................
    # Masking; these use the corresponding dantro data operations and work with
    # xarray.DataArray objects.
    "create_mask":                  create_mask,
    "where":                        where,
    # NOTE For applying a mask, use the xr.where operation.
    # xarray.pydata.org/en/stable/indexing.html#assigning-values-with-indexing

    # Array generation
    "populate_ndarray":             populate_ndarray,
    "build_object_array":           build_object_array,
    "expand_object_array":          expand_object_array,

    # extract labelling info, e.g. for creating higher-dimensional arrays
    "extract_dim_names":            extract_dim_names,
    "extract_coords_from_attrs":    extract_coords_from_attrs,

    # dantro-specific wrappers around other library's functionality
    "dantro.apply_along_axis":      apply_along_axis,
    "dantro.multi_concat":          multi_concat,
    "dantro.merge":                 merge,
    "dantro.expand_dims":           expand_dims,

    # coordinate transformations, working on shallow copies by default
    "transform_coords":             transform_coords,
    ".coords.transform":            transform_coords,

    # evaluating symbolic expressions using sympy
    "expression":                   expression,
    # NOTE: The `^` operator acts as XOR; use `**` for exponentiation!

    # numpy
    # CAUTION Some of these work in-place ...
    ".sum":                 lambda d, *a, **k: d.sum(*a, **k),
    ".prod":                lambda d, *a, **k: d.prod(*a, **k),
    ".cumsum":              lambda d, *a, **k: d.cumsum(*a, **k),
    ".cumprod":             lambda d, *a, **k: d.cumprod(*a, **k),

    ".mean":                lambda d, *a, **k: d.mean(*a, **k),
    ".std":                 lambda d, *a, **k: d.std(*a, **k),
    ".min":                 lambda d, *a, **k: d.min(*a, **k),
    ".max":                 lambda d, *a, **k: d.max(*a, **k),
    ".min.item":            lambda d, *a, **k: d.min(*a, **k).item(),
    ".max.item":            lambda d, *a, **k: d.max(*a, **k).item(),
    ".var":                 lambda d, *a, **k: d.var(*a, **k),
    ".argmin":              lambda d, *a, **k: d.argmin(*a, **k),
    ".argmax":              lambda d, *a, **k: d.argmax(*a, **k),
    ".argsort":             lambda d, *a, **k: d.argsort(*a, **k),
    ".argpartition":        lambda d, *a, **k: d.argpartition(*a, **k),

    ".transpose":           lambda d, *ax: d.transpose(*ax),
    ".squeeze":             lambda d, **k: d.squeeze(**k),
    ".flatten":             lambda d, **k: d.flatten(**k),
    ".diagonal":            lambda d, **k: d.diagonal(**k),
    ".trace":               lambda d, **k: d.trace(**k),
    ".sort":                lambda d, **k: d.sort(**k),
    ".fill":                lambda d, val: d.fill(val),
    ".round":               lambda d, **k: d.round(**k),
    ".take":                lambda d, i, **k: d.take(i, **k),
    ".swapaxes":            lambda d, a1, a2: d.swapaxes(a1, a2),
    ".reshape":             lambda d, s, **k: d.reshape(s, **k),
    ".astype":              lambda d, t, **k: d.astype(t, **k),

    "np.array":             np.array,
    "np.empty":             np.empty,
    "np.zeros":             np.zeros,
    "np.ones":              np.ones,
    "np.full":              np.full,
    "np.empty_like":        np.empty_like,
    "np.zeros_like":        np.zeros_like,
    "np.ones_like":         np.ones_like,
    "np.full_like":         np.full_like,

    "np.eye":               np.eye,
    "np.arange":            np.arange,
    "np.linspace":          np.linspace,
    "np.logspace":          np.logspace,

    "np.invert":            np.invert,
    "np.transpose":         np.transpose,
    "np.flip":              np.flip,
    "np.diff":              np.diff,
    "np.reshape":           np.reshape,
    "np.take":              np.take,
    "np.repeat":            np.repeat,
    "np.stack":             np.stack,
    "np.hstack":            np.hstack,
    "np.vstack":            np.vstack,
    "np.concatenate":       np.concatenate,

    "np.abs":               np.abs,
    "np.ceil":              np.ceil,
    "np.floor":             np.floor,
    "np.round":             np.round,
    "np.fmod":              np.fmod,

    "np.where":             np.where,
    "np.digitize":          np.digitize,
    "np.histogram":         np.histogram,
    "np.count_nonzero":     np.count_nonzero,

    "np.any":               np.any,
    "np.all":               np.all,
    "np.allclose":          np.allclose,
    "np.isnan":             np.isnan,
    "np.isclose":           np.isclose,
    "np.isinf":             np.isinf,
    "np.isfinite":          np.isfinite,
    "np.isnat":             np.isnat,
    "np.isneginf":          np.isneginf,
    "np.isposinf":          np.isposinf,
    "np.isreal":            np.isreal,
    "np.isscalar":          np.isscalar,

    "np.mean":              np.mean,
    "np.std":               np.std,
    "np.min":               np.min,
    "np.max":               np.max,
    "np.var":               np.var,
    "np.argmin":            np.argmin,
    "np.argmax":            np.argmax,

    # xarray
    ".sel":                 lambda d, *a, **k: d.sel(*a, **k),
    ".isel":                lambda d, *a, **k: d.isel(*a, **k),
    ".sel.item":            lambda d, *a, **k: d.sel(*a, **k).item(),
    ".isel.item":           lambda d, *a, **k: d.isel(*a, **k).item(),
    ".drop_sel":            lambda d, *a, **k: d.drop_sel(*a, **k),
    ".drop_isel":           lambda d, *a, **k: d.drop_isel(*a, **k),
    ".drop_dims":           lambda d, *a, **k: d.drop_dims(*a, **k),
    ".squeeze_with_drop":   lambda d, *a, **k: d.squeeze(*a, **k, drop=True),
    ".median":              lambda d, *a, **k: d.median(*a, **k),
    ".quantile":            lambda d, *a, **k: d.quantile(*a, **k),
    ".count":               lambda d, *a, **k: d.count(*a, **k),
    ".diff":                lambda d, *a, **k: d.diff(*a, **k),
    ".where":               lambda d, c, *a, **k: d.where(c, *a, **k),
    ".notnull":             lambda d, *a, **k: d.notnull(*a, **k),
    ".ffill":               lambda d, *a, **k: d.ffill(*a, **k),
    ".bfill":               lambda d, *a, **k: d.bfill(*a, **k),
    ".fillna":              lambda d, *a, **k: d.fillna(*a, **k),
    ".interpolate_na":      lambda d, *a, **k: d.interpolate_na(*a, **k),
    ".dropna":              lambda d, *a, **k: d.dropna(*a, **k),
    ".isin":                lambda d, *a, **k: d.isin(*a, **k),
    ".roll":                lambda d, *a, **k: d.roll(*a, **k),
    ".thin":                lambda d, *a, **k: d.thin(*a, **k),
    ".weighted":            lambda d, *a, **k: d.weighted(*a, **k),

    ".chunk":               lambda d, *a, **k: d.chunk(*a, **k),

    ".rolling":             lambda d, *a, **k: d.rolling(*a, **k),
    ".coarsen":             lambda d, *a, **k: d.coarsen(*a, **k),

    ".groupby":             lambda d, g, **k: d.groupby(g, **k),
    ".groupby_bins":        lambda d, g, **k: d.groupby_bins(g, **k),
    ".map":                 lambda ds, func, **k: ds.map(func, **k),
    ".reduce":              lambda d, func, **k: d.reduce(func, **k),

    ".rename":              lambda d, *a, **k: d.rename(*a, **k),
    ".expand_dims":         lambda d, *a, **k: d.expand_dims(*a, **k),
    ".swap_dims":           lambda d, *a, **k: d.swap_dims(*a, **k),
    ".assign_coords":       lambda d, *a, **k: d.assign_coords(*a, **k),
    ".assign_attrs":        lambda d, *a, **k: d.assign_attrs(*a, **k),
    ".assign":              lambda d, *a, **k: d.assign(*a, **k),

    ".to_dataframe":        lambda d, *a, **k: d.to_dataframe(*a, **k),

    ".to_array":            lambda ds, *a, **k: ds.to_array(*a, **k),
    ".rename_dims":         lambda ds, *a, **k: ds.rename_dims(*a, **k),
    ".rename_vars":         lambda ds, *a, **k: ds.rename_vars(*a, **k),
    ".drop_vars":           lambda ds, *a, **k: ds.drop_vars(*a, **k),
    ".assign_var":          lambda ds, name, var: ds.assign({name: var}),

    "xr.Dataset":           lambda *a, **k: xr.Dataset(*a, **k),
    "xr.DataArray":         lambda *a, **k: xr.DataArray(*a, **k),
    "xr.zeros_like":        lambda *a, **k: xr.zeros_like(*a, **k),
    "xr.ones_like":         lambda *a, **k: xr.ones_like(*a, **k),
    "xr.full_like":         lambda *a, **k: xr.full_like(*a, **k),

    "xr.merge":             lambda *a, **k: xr.merge(*a, **k),
    "xr.concat":            lambda *a, **k: xr.concat(*a, **k),
    "xr.align":             lambda *a, **k: xr.align(*a, **k),
    "xr.combine_nested":    lambda *a, **k: xr.combine_nested(*a, **k),
    "xr.combine_by_coords": lambda *a, **k: xr.combine_by_coords(*a, **k),

    # ... method calls that require additional Python packages
    ".rolling_exp":         lambda d, *a, **k: d.rolling_exp(*a, **k),
    ".rank":                lambda d, *a, **k: d.rank(*a, **k),

    # fitting with xr.DataArray.polyfit or scipy.optimize
    ".polyfit":             lambda d, *a, **k: d.polyfit(*a, **k),
    "curve_fit":
        lambda *a, **k: import_module_or_object("scipy.optimize",
                                                name="curve_fit")(*a, **k),
    # NOTE: Use the "lambda" operation to generate the callable
}) # End of default operation definitions
# fmt: on

# Add the boolean operators
_OPERATIONS.update(BOOLEAN_OPERATORS)



"""Implements operation hooks for the DAG parser implemented in
:py:mod:`~dantro._dag_utils`.
"""

import logging
from typing import Tuple

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

DAG_PARSER_OPERATION_HOOKS = dict()
"""Contains hooks that are invoked when a certain operation is parsed.
The values should be callables that receive ``operation, *args, **kwargs``
and return a 3-tuple of the manipulated ``operation, args, kwargs``.
The return values will be those that the Transformation object is created
from.

See the :ref:`dag_op_hooks` page for more information on integration and
available hooks.

Example of defining a hook and registering it:

.. testcode:: register_dag_parser_hook
    :hide:

    from typing import Tuple

.. testcode:: register_dag_parser_hook

    # Define hook function
    def _op_hook_my_operation(
        operation, *args, **kwargs
    ) -> Tuple[str, list, dict]:
        # ... do stuff here ...
        return operation, args, kwargs

    # Register with hooks registry
    from dantro.data_ops import DAG_PARSER_OPERATION_HOOKS

    DAG_PARSER_OPERATION_HOOKS["my_operation"] = _op_hook_my_operation

.. todo::

    Implement a decorator to automatically register operation hooks
"""


# -- Operation Hooks ----------------------------------------------------------
# NOTE:
#   - Names should follow ``op_hook_<operation-name>``
#   - A documentation entry should be added in doc/data_io/dag_op_hooks.rst


def op_hook_expression(operation, *args, **kwargs) -> Tuple[str, list, dict]:
    """An operation hook for the ``expression`` operation, attempting to
    auto-detect which symbols are specified in the given expression.
    From those, ``DAGTag`` objects are created, making it more convenient to
    specify an expression that is based on other DAG tags.

    The detected symbols are added to the ``kwargs.symbols``, if no symbol of
    the same name is already explicitly defined there.

    This hook accepts as positional arguments both the ``(expr,)`` form and
    the ``(prev_node, expr)`` form, making it more robust when the
    ``with_previous_result`` flag was set.

    If the expression contains the ``prev`` or ``previous_result`` symbols,
    the corresponding :py:class:`~dantro._dag_utils.DAGNode` will be added to
    the symbols additionally.

    For more information on operation hooks, see :ref:`dag_op_hooks`.
    """
    from sympy import Symbol
    from sympy.parsing.sympy_parser import parse_expr

    from .._dag_utils import DAGNode, DAGTag

    # Extract the expression string
    if len(args) == 1:
        expr = args[0]
    elif len(args) == 2:
        _, expr = args
    else:
        raise TypeError(
            f"Got unexpected positional arguments: {args}; expected either "
            "(expr,) or (prev_node, expr)."
        )

    # Try to extract all symbols from the expression
    all_symbols = parse_expr(expr, evaluate=False).atoms(Symbol)

    # Some symbols might already be given; only add those that were not given.
    # Also, convert the ``prev`` and ``previous_result`` symbols the
    # corresponding DAGNode object
    symbols = kwargs.get("symbols", {})
    for symbol in all_symbols:
        symbol = str(symbol)
        if symbol in symbols:
            log.remark(
                "Symbol '%s' was already specified explicitly! It "
                "will not be replaced.",
                symbol,
            )
            continue

        if symbol in ("prev", "previous_result"):
            symbols[symbol] = DAGNode(-1)
        else:
            symbols[symbol] = DAGTag(symbol)

    # For the case of missing ``symbols`` key, need to write it back to kwargs
    kwargs["symbols"] = symbols

    # For args, return _only_ ``expr``, as expected by the operation
    return operation, (expr,), kwargs


DAG_PARSER_OPERATION_HOOKS["expression"] = op_hook_expression



"""Implements operations that control the behaviour of the transformation or
pipeline in general, including functions that can be used for debugging."""

import logging
from typing import Any

from ..base import BaseDataContainer, BaseDataGroup
from ..exceptions import SkipPlot

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def raise_SkipPlot(
    cond: bool = True, *, reason: str = "", passthrough: Any = None
):
    """Raises :py:exc:`~dantro.exceptions.SkipPlot` to trigger that a plot is
    skipped without error, see :ref:`plot_mngr_skipping_plots`.

    If ``cond`` is False, this will do nothing but return the passthrough.

    Args:
        cond (bool, optional): Whether to actually raise the exception
        reason (str, optional): The reason for skipping, optional
        passthrough (Any, optional): A passthrough value which is returned if
            ``cond`` did not evaluate to True.
    """
    if cond:
        raise SkipPlot(reason)
    return passthrough


def print_data(
    data: Any, *, end: str = "\n", fstr: str = None, **fstr_kwargs
) -> Any:
    """Prints and passes on the data using :py:func:`print`.

    The print operation distinguishes between dantro types (in which case some
    more information is shown) and non-dantro types. If a custom format string
    is given, will always use that one.

    .. note::

        This is a passthrough-function: ``data`` is always returned without any
        changes. However, the print operation may lead to resolution of
        :py:mod:`~dantro.proxy` objects.

    Args:
        data (Any): The data to print
        end (str, optional): The ``end`` argument to the ``print`` call
        fstr (str, optional): If given, will use this to format the data for
            printing. The data will be the passed as first *positional*
            argument to the format string, thus addressable by ``{0:}`` or
            ``data`` (e.g. to access attributes via format-string syntax).
            If the format string is not ``None``, will *always* use the format
            string and not use the custom formatting for dantro objects.
        **fstr_kwargs: Keyword arguments passed to the :py:func:`format`
            call.

    Returns:
        Any: the given ``data``
    """
    if fstr is None and isinstance(data, BaseDataContainer):
        print(f"{data}, with data:\n{str(data.data)}\n", end=end)

    elif fstr is None and isinstance(data, BaseDataGroup):
        print(f"{data.tree}\n", end=end)

    else:
        fstr = fstr if fstr is not None else "{0:}"
        print(fstr.format(str(data), data=data, **fstr_kwargs), end=end)

    return data
